<?php

namespace Intergo\OptOut;

use Intergo\OptOut\Enums\OptOutEnum;
use Intergo\OptOut\Repositories\ContactRepository;
use Illuminate\Redis\Connections\Connection;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Redis;
use Intergo\Helpers\HashHelper;

abstract class AbstractOptOutService
{
    protected Redis $redis;
    protected DB $db;

    /**
     * AbstractOptOutService constructor.
     */
    public function __construct(Redis $redis, DB $db)
    {
        $this->redis = $redis;
        $this->db = $db;
    }

    /**
     * @return string
     */
    protected function cacheConnectionName()
    {
        return OptOutEnum::CACHE_CONNECTION;
    }

    /**
     * @return Connection
     */
    private function cacheConnection(): Connection
    {
        return $this->redis::connection(self::cacheConnectionName());
    }

    /**
     * @param OptOutVO $optOutVO
     * @return void
     */
    public function cacheOptOuts(OptOutVO $optOutVO)
    {
        $this->cacheGlobalOptOuts($optOutVO->authUserId, $optOutVO->appUserId);
        if (empty($optOutVO->listIds)) {
            return;
        }
        $this->cacheListOptOuts($optOutVO->authUserId, $optOutVO->listIds);
    }

    /**
     * @param $userId
     * @param $appUserId
     * @return void
     */
    public function cacheGlobalOptOuts($userId, $appUserId)
    {
        if ($this->existsOnCache($userId, OptOutEnum::GLOBAL_OPT_OUT)) {
            return;
        }
        $optOuts = (new ContactRepository($this->db))->getOptOutsByUserId($appUserId);
        //Cache result to avoid multiple query executions
        $this->setOnCache($userId, OptOutEnum::GLOBAL_OPT_OUT);
        foreach ($optOuts as $optOut) {
            $this->setOnCache($userId, $optOut->phone);
        }
    }

    /**
     * @param $userId
     * @param $listIds
     * @return void
     */
    public function cacheListOptOuts($userId, $listIds)
    {
        if ($this->existsOnCache($userId, OptOutEnum::LIST_OPT_OUT)) {
            return;
        }
        [$listIds, $listKey] = $this->processListIdsAndGetListKey($listIds);
        $optOuts = (new ContactRepository($this->db))->getOptOutsByListIds($listIds);
        //Cache result to avoid multiple query executions
        $this->setOnCache($userId, OptOutEnum::LIST_OPT_OUT);
        foreach ($optOuts as $optOut) {
            $this->setOnCache($userId, self::getListOptOutField($listKey, $optOut->phone));
        }
    }

    /**
     * Check if a number is unsubscribed
     *
     * @param $userId
     * @param $phoneNumber
     * @param $listIds
     * @param ?bool $bypassOptOut
     * @return bool
     */
    public function isUnsubscribed($userId, $phoneNumber, $listIds = null, ?bool $bypassOptOut = false): bool
    {
        if ($bypassOptOut) {
            return false;
        }
        if ($this->isGlobalOptOut($userId, $phoneNumber)) {
            return true;
        }
        if (empty($listIds)) {
            return false;
        }
        [$listIds, $listKey] = $this->processListIdsAndGetListKey($listIds);
        return $this->existsOnCache($userId, self::getListOptOutField($listKey, $phoneNumber));
    }

    /**
     * @param $userId
     * @param $phoneNumber
     * @return boolean
     */
    public function isGlobalOptOut($userId, $phoneNumber): bool
    {
        return $this->existsOnCache($userId, $phoneNumber);
    }

    /**
     * @param $listKey
     * @param $phoneNumber
     * @return string
     */
    public function getListOptOutField($listKey, $phoneNumber): string
    {
        return "$listKey:$phoneNumber";
    }

    /**
     * @param $listIds
     * @return array
     */
    public function processListIdsAndGetListKey($listIds): array
    {
        $listIds = is_array($listIds) ? $listIds : [$listIds];
        sort($listIds);
        $listKey = HashHelper::hash(implode(':', $listIds));
        return [$listIds, $listKey];
    }

    /**
     * @param $userId
     * @return void
     */
    public function invalidateOptOutFlags($userId)
    {
        $hashKey = OptOutEnum::cacheKey($userId);
        $fields = $this->cacheConnection()->hkeys($hashKey);
        foreach ($fields as $field) {
            $this->cacheConnection()->hdel($hashKey, $field);
        }
    }

    /**
     * @param $userId
     * @param string $type
     * @return string
     */
    public function key($userId, string $type): string
    {
        return "$userId:$type";
    }

    /**
     * @param string $userId
     * @param string $field
     * @return void
     */
    private function setOnCache(string $userId, string $field): void
    {
        $this->cacheConnection()->hset(OptOutEnum::cacheKey($userId), $field, true);
    }

    /**
     * @param $userId
     * @param $field
     * @return bool
     */
    private function existsOnCache($userId, $field): bool
    {
        return $this->cacheConnection()->hexists(OptOutEnum::cacheKey($userId), $field);
    }
}
